"""Top-level package for Melusine."""

__author__ = """Sacha Samama & Tom Stringer & Hugo Perrier"""
__email__ = 'ssamama@quantmetry.com, tstringer@quantmetry.com, hperrier@quantmetry.com'
__version__ = '1.8.7'
